﻿/*
 *  TOPPERS/ECNL Prototyping tool
 *
 *  Copyright (C) 2017 Cores Co., Ltd. Japan
 *
 *  上記著作権者は，以下の(1)～(4)の条件を満たす場合に限り，本ソフトウェ
 *  ア（本ソフトウェアを改変したものを含む．以下同じ）を使用・複製・改
 *  変・再配布（以下，利用と呼ぶ）することを無償で許諾する．
 *  (1) 本ソフトウェアをソースコードの形で利用する場合には，上記の著作
 *      権表示，この利用条件および下記の無保証規定が，そのままの形でソー
 *      スコード中に含まれていること．
 *  (2) 本ソフトウェアを，ライブラリ形式など，他のソフトウェア開発に使
 *      用できる形で再配布する場合には，再配布に伴うドキュメント（利用
 *      者マニュアルなど）に，上記の著作権表示，この利用条件および下記
 *      の無保証規定を掲載すること．
 *  (3) 本ソフトウェアを，機器に組み込むなど，他のソフトウェア開発に使
 *      用できない形で再配布する場合には，次のいずれかの条件を満たすこ
 *      と．
 *    (a) 再配布に伴うドキュメント（利用者マニュアルなど）に，上記の著
 *        作権表示，この利用条件および下記の無保証規定を掲載すること．
 *    (b) 再配布の形態を，別に定める方法によって，TOPPERSプロジェクトに
 *        報告すること．
 *  (4) 本ソフトウェアの利用により直接的または間接的に生じるいかなる損
 *      害からも，上記著作権者およびTOPPERSプロジェクトを免責すること．
 *      また，本ソフトウェアのユーザまたはエンドユーザからのいかなる理
 *      由に基づく請求からも，上記著作権者およびTOPPERSプロジェクトを
 *      免責すること．
 *
 *  本ソフトウェアは，無保証で提供されているものである．上記著作権者お
 *  よびTOPPERSプロジェクトは，本ソフトウェアに関して，特定の使用目的
 *  に対する適合性も含めて，いかなる保証も行わない．また，本ソフトウェ
 *  アの利用により直接的または間接的に生じたいかなる損害に関しても，そ
 *  の責任を負わない．
 *
 *  @(#) $Id$
 */
using System;
using System.Collections.Generic;
using System.Linq;
using Bridge;
using Bridge.Html5;

namespace WebMrbc
{
	public class Mruby
	{
		public object[] preRun = new object[0];
		public object[] postRun = new object[0];

		public Mruby()
		{
			/*canvas = Document.GetElementById("canvas");

			// As a default initial behavior, pop up an alert when webgl context is lost. To make your
			// application robust, you may want to override this behavior before shipping!
			// See http://www.khronos.org/registry/webgl/specs/latest/1.0/#5.15.2
			canvas.AddEventListener("webglcontextlost", new Action<Event>((Event e) => {
				Window.Alert("WebGL context lost. You will need to reload the page."); e.PreventDefault();
			}), false);*/
			setStatus = new Action<string>(_setStatus);
		}

		public void print(object arg)
		{
			string text;
			Array args;
			if (arg.GetType() == typeof(string))
				text = (string)arg;
			else if (arg.GetType().Is<Array>() && (args = (Array)arg).Length > 1) {
				var texts = new string[0];
				foreach (var ele in args) {
					texts.Push(ele.ToString());
				}
				text = texts.Join(" ");
			}
			else
				text = arg.ToString();

			App.WriteLine(text);
		}

		public void printErr(object arg)
		{
			string text;
			Array args;
			if (arg.GetType() == typeof(string))
				text = (string)arg;
			else if (arg.GetType().Is<Array>() && (args = (Array)arg).Length > 1) {
				var texts = new string[0];
				foreach (var ele in args) {
					texts.Push(ele.ToString());
				}
				text = texts.Join(" ");
			}
			else
				text = arg.ToString();

			App.WriteLine(text);
		}

		internal void run(string[] args)
		{
			App.Write("$ " + args.Join(" ") + "\r\n");

			var exe = (string)args.Shift();
			this["arguments"] = args;

			var onerror = Window.OnError;
			Window.OnError = new ErrorEventHandler((string message, string url, int lineNumber, int columnNumber, object error) => {
				var spinnerElement = Document.GetElementById("spinner");
				// TODO: do not warn on ok events like simulating an infinite loop or exitStatus
				setStatus("Exception thrown, see JavaScript console");
				spinnerElement.Style.Display = Display.None;
				setStatus = new Action<string>((text) => {
					if (!string.IsNullOrEmpty(text))
						printErr(new[] { "[post-exception status] " + text });
				});
				onerror(message, url, lineNumber, columnNumber, error);
				return false;
			});
			try {
				Script.Eval(exe + "(this)");
			}
			finally {
				Window.OnError = onerror;
			}
		}

		public HTMLElement canvas;
		public Action<string> setStatus;

		object _last;
		string _text;

		private void _setStatus(string text)
		{
			if (_last == null) _last = new { time = DateTime.Now.Ticks, text = "" };
			if (text == _text) return;
			_text = text;
			var progressElement = (HTMLProgressElement)Document.GetElementById("progress");
			var spinnerElement = (HTMLDivElement)Document.GetElementById("spinner");
			var m = text.Match(new RegExp(@"([^(]+)\((\d+(\.\d+)?)\/(\d+)\)"));
			var now = DateTime.Now.Ticks;
			if (m != null && now - DateTime.Now.Ticks < 30) return; // if this is a progress update, skip it if too soon
			if (m != null) {
				text = m[1];
				progressElement.Value = Script.ParseInt(m[2]) * 100;
				progressElement.Max = Script.ParseInt(m[4]) * 100;
				progressElement.SetAttribute("hidden", "false");
				spinnerElement.SetAttribute("hidden", "false");
			}
			else {
				progressElement.Value = 0.0;
				progressElement.Max = 0.0;
				progressElement.SetAttribute("hidden", "true");
				if (!string.IsNullOrEmpty(text))
					spinnerElement.Style.Display = Display.None;
			}
			var statusElement = (HTMLDivElement)Document.GetElementById("status");
			statusElement.InnerHTML = text;
		}

		public int totalDependencies = 0;

		public void monitorRunDependencies(int left)
		{
			totalDependencies = System.Math.Max(totalDependencies, left);
			setStatus((left != 0) ? "Preparing... (" + (totalDependencies - left) + "/" + totalDependencies + ")" : "All downloads complete.");
		}

		[External, Convention(Notation.PascalCase)]
		internal string UTF8ArrayToString(Uint8Array buf, int index)
		{
			throw new NotImplementedException();
		}

		[External]
		internal string intArrayToString(Uint8Array buf)
		{
			throw new NotImplementedException();
		}

		[External]
		internal int stringToUTF8Array(string str, Uint8Array outU8Array, int outIdx, int maxBytesToWrite)
		{
			throw new NotImplementedException();
		}

		[External]
		internal int lengthBytesUTF8(string str)
		{
			throw new NotImplementedException();
		}

		[Convention(Notation.PascalCase)]
		internal Uint8Array UTF8StringToArray(string str)
		{
			var len = lengthBytesUTF8(str);
			var result = new Uint8Array(len);
			stringToUTF8Array(str, result, 0, len);
			return result;
		}

		[Script("return this.FS;")]
		internal FileSystem GetFileSystem()
		{
			return (FileSystem)this["FS"];
		}

		[External]
		internal void stdin(object data)
		{
			throw new NotImplementedException();
		}
	}
}
